{*****************************************************************************
** Image unit,                                                             **
**   by Steven Don                                                         **
*****************************************************************************
** A unit to load a variety of image formats to a buffer.                  **
**                                                                         **
** Supported image formats are: Windows BMP                                **
**                              Compuserve GIF (87a and 89a)               **
**                              Z-Soft PCX                                 **
**                                                                         **
** Images must be in 256 colours and may be up to 320x200 in size.         **
*****************************************************************************
** Three functions are available to the calling program:                   **
**                                                                         **
**   LoadBMP (Name, Location);                                             **
**   LoadGIF (Name, Location);                                             **
**   LoadPCX (Name, Location);                                             **
**                                                                         **
** Name is the filename, including extension (e.g. TEST.GIF)               **
** Location is a pointer to an Array [0..64000] of Byte; where the image   **
** data will be stored.                                                    **
** All functions are boolean. If they return false, check IMGERROR to find **
** out what happened.                                                      **
**                                                                         **
** For questions, feel free to e-mail me.                                  **
**                                                                         **
**    shd@earthling.net                                                    **
**    http://shd.cjb.net                                                   **
**                                                                         **
****************************************************************************}
unit IMGUnit;

interface

{Error codes}
const
  FileNotFound      = $01;
  UnsupportedFormat = $02;
  TooLarge          = $03;

var
  IMGError : Byte;

{Functions available to the calling program}
function LoadBMP (FileName : String; Location : Pointer) : Boolean;
function LoadPCX (FileName : String; Location : Pointer) : Boolean;
function LoadGIF (FileName : String; Location : Pointer) : Boolean;

implementation

type
  BufferType         = Array [0..63999] of Byte;
  BufferPointer      = ^BufferType;

  {Windows Bitmap header}
  BMPHeader = record
    bfType           : Word;
    bfSize,
    bfReserved,
    bfOffBits,
    biSize,
    biWidth,
    biHeight         : LongInt;
    biPlanes,
    biBitCount       : Word;
    biCompression,
    biSizeImage,
    biXPelsPerMeter,
    biYPelsPerMeter,
    biClrUsed,
    biClrImportant   : LongInt;
  end;

  {Z-Soft PCX header}
  PCXHeader = record
    Manufacturer,
    Version,
    Encoding,
    BitsPerPixel     : Byte;
    xMin,
    yMin,
    xMax,
    yMax             : Integer;
    Other            : Array [1..116] of Byte;
  end;

  {Compuserve GIF header}
  GIFHeader = record
    Signature        : String [6];
    ScreenWidth,
    ScreenHeight     : Word;
    Depth,
    Background,
    Zero             : Byte;
  end;

  {and GIF image descriptor}
  GIFDescriptor = record
    Separator        : Char;
    ImageLeft,
    ImageTop,
    ImageWidth,
    ImageHeight      : Word;
    Depth            : Byte;
  end;

var
  {The buffer to draw to}
  Buffer             : ^BufferType;

{Function prototypes}
procedure SetDAC (DAC, R, G, B : Byte); forward;
procedure PutPixel (X, Y : Word; C : Byte); forward;

{This loads a BMP File to the specified location.}
function LoadBMP;
var
  BMPFile                        : File;
  Header                         : BMPHeader;
  Palette                        : Array [0..255, 0..3] of Byte;
  OffSet, Lines, Counter         : Word;

begin
  {No errors yet}
  IMGError := 0;

  {Check to see whether the file exists and can be opened}
  {$I-}
  Assign (BMPFile, FileName);
  Reset (BMPFile, 1);
  if IOResult <> 0 then begin
    IMGError := FileNotFound;
    LoadBMP := false;
    Exit;
  end;
  {$I+}

  {Read in header information}
  BlockRead (BMPFile, Header, SizeOf (Header));

  {Check to see whether we can display it}
  if (Header.bfType <> 19778)
  or (Header.bfReserved <> 0)
  or (Header.biPlanes <> 1)
  or (Header.biCompression <> 0)
  or (Header.biBitCount <> 8) then begin
    Close (BMPFile);
    IMGError := UnsupportedFormat;
    LoadBMP := false;
    Exit;
  end;

  if (Header.biWidth > 320) or (Header.biHeight > 200) then begin
    Close (BMPFile);
    IMGError := TooLarge;
    LoadBMP := false;
    Exit;
  end;

  {Load in the palette}
  BlockRead (BMPFile, Palette, 1024);

  {Set the palette}
  for Counter := 0 to 255 do begin
    SetDAC (Counter,
            Palette [Counter, 2] shr 2,
            Palette [Counter, 1] shr 2,
            Palette [Counter, 0] shr 2);
  end;

  {Figure out where to load graphics data}
  Offset := (Header.biHeight - 1) * 320;
  Buffer := Location;

  {We've read no lines so far}
  Lines := 0;

  {Stored upside down, so moving up}
  while (Lines < Header.biHeight) do begin
    {Read next line}
    BlockRead (BMPFile, Buffer^ [Offset], Header.biWidth);
    {Move up one line on the screen}
    dec (offset, 320);
    {increase amount of lines read}
    inc (lines);
  end;

  {Close the file}
  Close (BMPFile);

  {Successful}
  LoadBMP := true;
end;

{This loads a BMP File to the specified location.}
function LoadPCX;
var
  PCXFile                        : File;
  Header                         : PCXHeader;
  Palette                        : Array [0..255,0..2] of Byte;
  PCXLen                         : LongInt;
  DOffSet, OffSet, Width, Height : Word;
  HowMany, DataByte, Counter     : Byte;
  PCXBuffer                      : Array [1..512] of Byte;
  nextPointer                    : Word;
  X, Y                           : Integer;

  {This function reads the next byte from the PCX file}
  function NextByte : Byte;
  begin
    {Check to see whether next block should be read}
    if nextPointer < 1 then begin
      {If so, read it and reset the pointer}
      {$I-}
      BlockRead (PCXFile, PCXBuffer [1], SizeOf (PCXBuffer));
      if IOResult <> 0 then;
      {$I+}
      nextPointer := 1;
    end;

    {Return byte in buffer}
    NextByte := PCXBuffer [nextPointer];

    {Adjust pointer}
    inc (nextPointer);
    if nextPointer > 512 then NextPointer := 0;
  end;

begin
  {No errors yet}
  IMGError := 0;

  {Check to see whether the file exists and can be opened}
  {$I-}
  Assign (PCXFile, FileName);
  Reset (PCXFile, 1);
  if IOResult <> 0 then begin
    IMGError := FileNotFound;
    LoadPCX := false;
    Exit;
  end;
  {$I+}

  {Read in header information}
  BlockRead (PCXFile, Header, SizeOf (Header));

  {Check to see whether we can display it}
  if Header.Version <> 5 then begin
    Close (PCXFile);
    IMGError := UnsupportedFormat;
    LoadPCX := false;
    Exit;
  end;

  if (Header.xMax > 319) or (Header.xMin < 0)
  or (Header.yMax > 199) or (Header.yMin < 0) then begin
    Close (PCXFile);
    IMGError := TooLarge;
    LoadPCX := false;
    Exit;
  end;

  {Load in the palette}
  Seek (PCXFile, FileSize (PCXFile) - 769);
  {Read in identifier}
  BlockRead (PCXFile, DataByte, 1);
  if DataByte <> 12 then begin
    Close (PCXFile);
    IMGError := UnsupportedFormat;
    LoadPCX := false;
    Exit;
  end;

  BlockRead (PCXFile, Palette[0, 0], 768);

  {Set the palette}
  for Counter := 0 to 255 do begin
    SetDAC(Counter,
           Palette [Counter, 0] shr 2,
           Palette [Counter, 1] shr 2,
           Palette [Counter, 2] shr 2);
  end;

  {Go back to start of graphic data}
  Seek (PCXFile, 128);

  x := Header.xMin;
  y := Header.yMin;
  inc (Header.xMax);
  inc (Header.yMax);

  {Figure out where to load graphics data}
  Buffer := Location;

  {Initialize loader}
  nextPointer := 0;

  {Decode and display graphics}
  While (Y < Header.yMax) do begin
    {Read next byte}
    DataByte := NextByte;

    {Reset counter}
    HowMany := 1;

    {If it is encoded, extract the count information and read in the colour byte}
    if (DataByte and $C0) = $C0 THEN begin
      HowMany := DataByte and $3F;
      DataByte := NextByte;
    end;

    {Display it}
    for Counter := 1 to HowMany do begin
      {Store pixel in the buffer}
      PutPixel (X, Y, DataByte);
      {Next pixel}
      inc (X);

      {If End of Line reached, next line}
      if X = Header.xMax then begin
        inc (Y);
        X := Header.xMin;
      end;

    end;
  end;

  {Close the file}
  Close (PCXFile);

  {Successful}
  LoadPCX := true;
end;

{This loads a GIF File to the specified location.}
function LoadGIF;
var
  {For loading from the GIF file}
  GIFFile      : File;
  Header       : GIFHeader;
  Descriptor   : GIFDescriptor;
  Temp         : Byte;
  BPointer     : Word;
  Buffer       : Array [0..256] of Byte;

  {Colour information}
  BitsPerPixel,
  NumOfColours,
  DAC          : Byte;
  Palette      : Array [0..255, 0..2] of Byte;

  {Coordinates}
  X, Y,
  tlX, tlY,
  brX, brY     : Word;

  {GIF data is stored in blocks of a certain size}
  BlockSize    : Byte;

  {The string table}
  Prefix,
  Suffix       : Array [0..4096] Of Word;
  OutCode      : Array [0..1024] Of Byte;
  FirstFree,
  FreeCode     : Word;

  {All the code information}
  InitCodeSize,
  CodeSize     : Byte;
  Code,
  OldCode,
  MaxCode      : Word;

  {Special codes}
  ClearCode,
  EOICode      : Word;

  {Used while reading the codes}
  BitsIn       : Byte;

  {Local function to read from the buffer}
  function LoadByte : Byte;
  begin
    {Read next block}
    if (BPointer = BlockSize) then begin
      {$I-}
      BlockRead (GIFFile, Buffer, BlockSize + 1);
      if IOResult <> 0 then;
      {$I+}
      BPointer := 0;
    end;
    {Return byte}
    LoadByte := Buffer [BPointer];
    inc (BPointer);
  end;

  {Local procedure to read the next code from the file}
  procedure ReadCode;
  var
    Counter : Integer;

  begin
    Code := 0;
    {Read the code, bit by bit}
    for Counter := 0 To CodeSize - 1 do begin
      {Next bit}
      inc (BitsIn);

      {Maybe, a new byte needs to be loaded with a further 8 bits}
      if (BitsIn = 9) then begin
        Temp := LoadByte;
        BitsIn := 1;
      end;

      {Add the current bit to the code}
      if ((Temp and 1) > 0) then inc (Code, 1 shl Counter);
      Temp := Temp shr 1;
    end;
  end;

  {Local procedure to draw a pixel}
  procedure NextPixel (c : Word);
  begin
    {Actually draw the pixel on screen}
    PutPixel (X, Y, c and 255);

    {Move on to next pixel}
    inc (X);

    {Or next row, if necessary}
    if (X = brX) then begin
      X := tlX;
      inc (Y);
    end;
  end;

  {Local function to output a string. Returns the first character.}
  function OutString (CurCode : Word) : Byte;
  var
    OutCount : Word;

  begin
    {If it's a single character, output that}
    if CurCode < 256 then begin
      NextPixel (CurCode);
    end else begin
      OutCount := 0;

      {Store the string, which ends up in reverse order}
      repeat
        OutCode [OutCount] := Suffix [CurCode];
        inc (OutCount);
        CurCode := Prefix [CurCode];
      until (CurCode < 256);

      {Add the last character}
      OutCode [OutCount] := CurCode;
      inc (OutCount);

      {Output all the string, in the correct order}
      repeat
        dec (OutCount);
        NextPixel (OutCode [OutCount]);
      until OutCount = 0;
    end;
    {Return 1st character}
    OutString := CurCode;
  end;

begin
  {No errors yet}
  IMGError := 0;

  {Check to see whether the file exists and can be opened}
  {$I-}
  Assign (GIFFile, FileName);
  Reset (GIFFile, 1);
  if IOResult <> 0 then begin
    IMGError := FileNotFound;
    LoadGIF := false;
    Exit;
  end;
  {$I+}

  {Read header}
  Header.Signature [0] := Chr (6);
  Blockread (GIFFile, Header.Signature [1], sizeof (Header) - 1);

  {Check signature and terminator}
  if ((Header.Signature <> 'GIF87a') and (Header.Signature <> 'GIF89a'))
  or (Header.Zero <> 0) then begin
    Close (GIFFile);
    IMGError := UnsupportedFormat;
    LoadGIF := false;
    Exit;
  end;

  {Get amount of colours in image}
  BitsPerPixel := 1 + (Header.Depth and 7);
  NumOfColours := (1 shl BitsPerPixel) - 1;

  {Load global colour map}
  BlockRead (GIFFile, Palette, 3 * (NumOfColours + 1));
  for DAC := 0 to NumOfColours do begin
    SetDAC(DAC, Palette [DAC, 0] shr 2,
                Palette [DAC, 1] shr 2,
                Palette [DAC, 2] shr 2);
  end;

  {Load the image descriptor}
  BlockRead (GIFFile, Descriptor, sizeof (Descriptor));

  if (Descriptor.Separator <> ',')
  or (Descriptor.Depth and 192 > 0) then begin
    Close (GIFFile);
    IMGError := UnsupportedFormat;
    LoadGIF := false;
    Exit;
  end;

  {Get image corner coordinates}
  tlX := Descriptor.ImageLeft;
  tlY := Descriptor.ImageTop;
  brX := tlX + Descriptor.ImageWidth;
  brY := tlY + Descriptor.ImageHeight;

  {Get initial code size}
  BlockRead (GIFFile, CodeSize, 1);

  {GIF data is stored in blocks, so it's necessary to know the size}
  BlockRead (GIFFile, BlockSize, 1);

  {Start loader}
  BPointer := BlockSize;

  {Special codes used in the GIF spec}
  ClearCode        := 1 shl CodeSize;    {Code to reset}
  EOICode          := ClearCode + 1;     {End of file}

  {Initialize the string table}
  FirstFree        := ClearCode + 2;     {Strings start here}
  FreeCode         := FirstFree;         {Strings can be added here}

  {Initial size of the code and its maximum value}
  inc (CodeSize);
  InitCodeSize     := CodeSize;
  MaxCode          := 1 shl CodeSize;

  BitsIn := 8;

  {Start at top left of image}
  X := Descriptor.ImageLeft;
  Y := Descriptor.ImageTop;

  repeat
    {Read next code}
    ReadCode;

    {If it's an End-Of-Information code, stop processing}
         if Code = EOICode then break
    {If it's a clear code...}
    else if Code = ClearCode then begin
      {Clear the string table}
      FreeCode := FirstFree;

      {Set the code size to initial values}
      CodeSize := InitCodeSize;
      MaxCode  := 1 shl CodeSize;

      {The next code may be read}
      ReadCode;
      OldCode := Code;

      {Set pixel}
      NextPixel (Code);
    {Other codes}
    end else begin
      {If the code is already in the string table, it's string is displayed,
      and the old string followed by the new string's first character is
      added to the string table.}
      if (Code < FreeCode) then
        Suffix [FreeCode] := OutString (Code)
      else begin
      {If it is not already in the string table, the old string followed by
      the old string's first character is added to the string table and
      displayed.}
        Suffix [FreeCode] := OutString (OldCode);
        NextPixel (Suffix [FreeCode]);
      end;

      {Finish adding to string table}
      Prefix [FreeCode] := OldCode;
      inc (FreeCode);

      {If the code size needs to be adjusted, do so}
      if (FreeCode >= MaxCode) and (CodeSize < 12) then begin
        inc (Codesize);
        MaxCode := MaxCode shl 1;
      end;

      {The current code is now old}
      OldCode := Code;
    end;

  until Code = EOICode;

  {Close the GIF file}
  Close (GIFFile);
end;


{This sets a DAC register to a specific Red Green Blue-value}
procedure SetDAC;
begin
  Port[$3C8] := DAC;
  Port[$3C9] := R;
  Port[$3C9] := G;
  Port[$3C9] := B;
end;

{Sets a pixel in the buffer to a specific colour}
procedure PutPixel;
begin
  Buffer^ [Y shl 8 + Y shl 6 + X] := C;
end;

end.